#
# Licence GPL v3.
#
# Copyright Chris Johns (chrisj@rtems.org)
#
import linecache
import optparse
import os
import pqconfig
import re
import sys
import tailer
import time

class error:
    def __init__(self, target = None, line = None):
        self.clear()
        if line:
            self.parse(target, line)

    def __str__(self):
        return 'type=' + str(self.type) + ', file=' + str(self.file) + \
            ', ftype=' + str(self.ftype) + ', line=' + str(self.line) + \
            ', info=' + str(self.info) + ', source=' + str(self.source) + \
            ', error=' + str(self.error)

    def clear(self):
        self.state = 'empty'
        self.file = None
        self.type = None
        self.ftype = None
        self.line = None
        self.info = None
        self.source = None
        self.error = None
        
    def valid(self):
        return self.state != 'empty'

    def parse(self, target, line):
        self.clear()
        se = line.split(':')
        if len(se) < 3:
            self.clear()
            return
        # Sometimes an extra number can be added to some gcc warnings
        try:
            num = int(se[2])
            self.type = se[3].strip()
            self.info = ':'.join(se[4:])
        except ValueError, e:
            self.info = se[len(se) - 1].strip()
            if len(se) == 3 and self.info.startswith('undefined reference to'):
                self.type = 'error'
            else:
                self.type = se[2].strip()
        # If an ICE change it to error
        if self.type == 'internal compiler error':
            self.info = self.type + ' ' + se[3].strip() + '|' + se[4].strip()
            self.type = 'error'
        # If a linker relocation type error
        if self.type == 'relocation truncated to fit':
            self.type = 'error'
        # See of the file is binary, source, or stdin ?
        self.ftype = 'source'
        self.line = 0
        self.file = se[0]
        if self.file == '{standard input}':
            self.ftype = 'stdin'
        else:
            if self.file.endswith(os.path.join('bin', 'ld')):
                self.ftype = 'binary'
                self.line = 0
                self.type = se[1].strip()
                self.info = se[len(se) - 1].strip()
            else:
                try:
                    self.line = int(se[1])
                except ValueError, e:
                    self.clear()
                    return
        self.state = 'current'
        self.source = None
        self.error = self.file + ':' + ':'.join(se[1:])
        self.target = target
        
    def get_source(self, source, window = 10):
        file = os.path.join(source, self.file)
        if self.valid() and os.path.isfile(file):
            line = self.line
            if line < (window / 2):
                line = 1
            else:
                line -= window / 2
            self.source = []
            for l in range(line, line + window):
                sl = linecache.getline(file, l)
                self.source.append((l, sl[:-1]))

class errors:
    def __init__(self, base = None, data = None, source = None, source_find = None):
        self.stderr = {}
        self.system = {}
        self.user = {}
        self.user_dirty = False
        self.baseline = {}
        self.baseline_dirty = False
        self.fix = {}
        self.fix_dirty = False
        self.dup_count = 0
        self.counts = { 'error': 0, 'warning':0, 'note':0 }
        self.base = base
        self.data = data
        self.source = source
        self.source_find = source_find
        if base:
            self.load_system(base)
        if data:
            self.load_user(data)
            self.load_baseline(data)
            self.load_fix(data)

    def __del__(self):
        if self.data:
            self.unload_user(self.data)
            self.unload_baseline(self.data)
            self.unload_fix(self.data)
        
    def walk_stderr(self):
        def cmp(k):
            if self.stderr[k]['error'].type == 'error':
                return 100000000
            return self.stderr[k]['dup']
        for key in sorted(self.stderr.iterkeys(), key = cmp, reverse = True):
            yield self.stderr[key]

    def add(self, to, target, key):
        e = error(target, key)
        if e.valid():
            if to == 'baseline':
                self.baseline[key] = { 'dup': 0, 'error': e }
                self.baseline_dirty = True
            elif to == 'user':
                self.user[key] = { 'dup': 0, 'error': e }
                self.user_dirty = True
            elif to == 'fix':
                self.fix[key] = { 'dup': 0, 'error': e }
                self.fix_dirty = True

    def duplicates(self):
        return self.dup_count

    def count(self):
        return len(self.stderr)

    def check(self, target, key, map):
        if key in map:
            t = map[key]['error'].target
            if t == 'all' or t == target:
                return True
        return False

    def check_all(self, target, key, baseline = True):
        if key[-1:] != ':' and (self.check(target, key, self.system) or \
                (baseline and self.check(target, key, self.baseline)) or \
                self.check(target, key, self.user)):
            return True
        return False

    def load_stderr(self, target, stderr):
        stderr.seek(0)
        count = 0
        linker_error_line = 0
        linker_error_length = 0
        for line in stderr:
            count += 1
            if len(line) == 0:
                continue
            if line[-1:] == os.linesep:
                line = line[:-1]
            # check for a linker error
            # first look for 2 types, the 3 lines long and the 2 line version
            if linker_error_line == 0 and re.compile(".*\(.*\):.*").search(line, 1):
                if line[-1:] == ':':
                    linker_error_length = 3
                else:
                    linker_error_length = 2
                linker_error_line = 1
                continue
            if linker_error_line == 1 and linker_error_length == 3:
                linker_error_line += 1
                sl = line.split(':')
                if len(sl) + 2:
                    try:
                        num = int(sl[1])
                        sl.insert(2, ' error')
                        line = ':'.join(sl)
                    except ValueError, e:
                        pass
            elif linker_error_line >= 1:
                linker_error_line = 0
                continue
            if line.find(os.path.join('bin', 'ld')) >= 0:
                sl = line.split(':')
                if len(sl) == 2:
                    if sl[1].find('is not within region') >= 0:
                        sl.insert(1, ' error')
                        line = ':'.join(sl)
            line = normalise_error(line, target, self.source, self.source_find)
            try:
                if not self.check_all(target, line):
                    if line in self.stderr:
                        self.stderr[line]['dup'] += 1
                        self.dup_count += 1
                    else:
                        e = error(target, line)
                        if e.valid():
                            if e.type not in self.counts:
                                self.counts[e.type] = 1
                            else:
                                self.counts[e.type] += 1
                            self.stderr[e.error] = { 'dup': 0, 'error': e }
            except:
                print "errors: load failed:", stderr, ":" , count, ":", line
                raise

    def load_external(self, name, map, *dirty):
        try:
            file = open(name)
        except IOError, e:
            return False
        for rec in file:
            recs = rec.split('\t')
            if len(recs) == 2:
                e = error(recs[0], recs[1])
                map[recs[1]] = { 'dup': 0, 'error': e }
        file.close()
        if dirty:
            dirty = False
        return True

    def unload_external(self, name, map, dirty):
        if dirty:
            try:
                file = open(name, 'w')
            except IOError, e:
                return False
            for key in map:
                file.writelines('%s\t%s' % (map[key]['error'].target, key))
            file.close()

    def load_system(self, base):
        return self.load_external(os.path.join(base, 'pq', 'pq-rtems-errors'),
                                  self.system)

    def load_user(self, data):
        return self.load_external(os.path.join(data, '.pq-rtems-errors'),
                                  self.user, self.user_dirty)

    def load_baseline(self, data):
        return self.load_external(os.path.join(data, '.pq-rtems-baseline'),
                                  self.baseline, self.baseline_dirty)

    def load_fix(self, data):
        return self.load_external(os.path.join(data, '.pq-rtems-fix'),
                                  self.fix, self.fix_dirty)

    def unload_user(self, data):
        return self.unload_external(os.path.join(data, '.pq-rtems-errors'),
                                    self.user, self.user_dirty)

    def unload_baseline(self, data):
        return self.unload_external(os.path.join(data, '.pq-rtems-baseline'),
                                    self.baseline, self.baseline_dirty)

    def unload_fix(self, data):
        return self.unload_external(os.path.join(data, '.pq-rtems-fix'),
                                    self.fix, self.fix_dirty)

class history:
    def __init__(self, data = None, output = None):
        self.history = None
        self.data = data
        self.output = output
        if self.data:
            self.load(self.data)

    def __del__(self):
        if self.data:
            self.unload(self.data)
    
    def sort(self):
        def cmp(h):
            try:
                t = time.strptime(h[1], '%d %b %Y')
            except:
                t = 0
            return t
        if self.history:
            for h in range(0, len(self.history)):
                for f in range(0, len(self.history[h])):
                    if self.history[h][f] == None:
                        self.history[h][f] = ''
            self.history.sort(key = cmp, reverse = True)
    
    def add(self, workspace, target, errs, first_page, t):
        if not self.history:
            self.history = []
        date = time.strftime('%d %b %Y', time.localtime(t))
        updated = False
        for h in range(0, len(self.history)):
            if self.history[h][0] == target and self.history[h][1] == date:
                self.history[h][2] = str(errs.counts['error'])
                self.history[h][3] = str(errs.counts['warning'])
                self.history[h][4] = str(errs.counts['note'])
                self.history[h][5] = str(errs.duplicates())
                self.history[h][6] = status(workspace, target)
                self.history[h][7] = first_page
                updated = True
                break
        if not updated:
            self.history.append([target, date,
                                 str(errs.counts['error']),
                                 str(errs.counts['warning']),
                                 str(errs.counts['note']),
                                 str(errs.duplicates()),
                                 status(workspace, target),
                                 first_page])
        self.sort()

    def remove_oldest(self, target):
        if self.history:
            self.sort()
            last = None
            count = 0
            for h in range(0, len(self.history)):
                if self.history[h][0] == target:
                    if count < pqconfig.history_records():
                        count += 1
                    else:
                        last = h
            if last:
                print self.history[last]
                if self.output:
                    t = time.strptime(self.history[last][1], '%d %b %Y')
                    name = pqconfig.publish_filename(self.history[last][0], t)
                    od = os.listdir(self.output)
                    for f in od:
                        if os.path.basename(f).startswith(name):
                            os.remove(os.path.join(self.output, f))
                self.history.remove(self.history[last])

    def load(self, data):
        self.history = []
        try:
            file = open(os.path.join(data, '.pq-stats-history'))
            for line in file:
                self.history.append(line[:-1].split(':'))
            file.close()
        except IOError:
            pass
        self.sort()

    def unload(self, data):
        if self.history:
            self.sort()
            file = None
            try:
                file = open(os.path.join(data, '.pq-stats-history'), 'w')
                for h in self.history:
                    file.writelines(':'.join(h) + os.linesep)
                file.close()
            except IOError:
                if file:
                    file.close()
            self.history = None

def string_count(file, strings):
    file.seek(0)
    counts = []
    for c in range(0, len(strings)):
        counts.append(0)
    for line in file:
        for s in range(0, len(strings)):
            if line.find(strings[s]) > 0:
                counts[s] += 1
    return counts

def normalise_error(err, target, source, source_find):
    pos = err.find(':')
    if pos >= 0:
        file = err[:pos]
        if source:
            file = os.path.join(source, file)
        file = os.path.normpath(file)
        if source_find and not os.path.isfile(file):
            file = source_find(file, target, source)
        if source:
            file = file.replace(os.path.join(source, ''), '') 
        err = file + err[pos:]
    return err

def status(workspace, target):
    try:
        what = 'UNKNOWN'
        stderr = open(os.path.join(workspace, target, 'stderr.txt'))
        end = tailer.tail(stderr, 1)
        stderr.close()
        if len(end) == 1:
            if end[0].startswith('terminate:'):
                what = 'DIED'
            elif end[0].startswith('failure:'):
                what = 'FAILED'
            else:
                what = 'PASSED'
    except IOError, e:
        pass
    return what

def stats(base, data, workspace, source, target):
    es = errors(base, data, source)
    try:
        stderr = open(os.path.join(workspace, target, 'stderr.txt'))
        es.load_stderr('none', stderr)
        stderr.close()
        s = 'Errors: %3d Warnings: %5d Notes: %4d Duplicates: %5d' % \
            (es.counts['error'], es.counts['warning'], es.counts['note'], es.duplicates())
        del es
        return s
    except IOError, e:
        pass
    return 'No error file located'

def target_name(target, width = 10):
    return "%-*s" % (width, target.upper())

def target_list(workspace):
    targets = []
    for f in os.listdir(workspace):
        fp = os.path.join(workspace, f)
        if os.path.isdir(fp):
            if os.path.isfile(os.path.join(fp, 'stderr.txt')):
                targets.append(os.path.basename(fp))
    targets.sort()
    return targets

def error_source_print(err, dups, source):
    print '%s:%d: %s' % (err.type.upper(), dups, err.info)
    print ' %s: %s:%d' % (target_name(err.target, 1), err.file, err.line)
    if err.ftype == 'source':
        err.get_source(source)
        if err.source:
            for s in err.source:
                if err.line == s[0]:
                    c1 = '*'
                    c2 = '>'

                    c1 = ' '
                    c2 = '|'
                print '%c%4d%c%s' % (c1, s[0], c2, s[1])

def parse_std_args(argv, opts):
    cmd_opts = optparse.OptionParser(option_list = opts)
    (options, args) = cmd_opts.parse_args(args=argv)
    config = options.config
    data = options.data
    if options.workspace:
        workspace = options.workspace
    else:
        workspace = os.path.abspath('.')
    if options.targets:
        targets = options.targets.split(',')
    else:
        targets = target_list(workspace)
    if options.source:
        source = options.source
    else:
        source = os.path.join(workspace, 'src')
    if not os.path.isdir(workspace):
        print >> sys.stderr, "workspace not a directory:", workspace
        sys.exit(1)
    return (options, config, data, workspace, source, targets)

# Regular expression to find quotes.
qstr = re.compile('[rR]?\'([^\\n\'\\\\]|\\\\.)*\'|[rR]?"([^\\n"\\\\]|\\\\.)*"')

def arg_list(args):
    """Turn a string of arguments into a list suitable for
    spawning a command. If the args are already a list return
    it."""
    if type(args) is list:
        return args
    argstr = args
    args = []
    while len(argstr):
        qs = qstr.search(argstr)
        if not qs:
            args.extend(argstr.split())
            argstr= ''
        else:
            # We have a quoted string. Get the string before
            # the quoted string and spltt on white space then
            # add the quoted string as an option then remove
            # the first + quoted string and try again
            front = argstr[:qs.start()]
            args.extend(front.split())
            args.append(argstr[qs.start() + 1:qs.end() - 1])
            argstr = argstr[qs.end():]
    return args        

def command(cmd, dir, stdout, stderr):
    if not os.path.isdir(dir):
        print >> sys.stderr, "cannot find directory:", dir
        sys.exit(1)
    process = subprocess.Popen(arg_list(cmd), cwd = dir,
                               stdout = stdout, stderr = stderr)
    
